/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on 23.02.2004
 */
package at.spardat.properties;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

/**
 * extended version of java.util.Properties which supports:
 * <li>variable replacement
 * <li>include
 * <li>property attributes
 * <li>property conditions
 * 
 * Objects of this class are for internal use of the property API only.
 * 
 * @author Martin Kahr
 */
public class ExtendedProperties extends Properties {
	private PropertyData data;	
	
	public ExtendedProperties() {
		this(ExtendedProperties.class.getClassLoader());
	}
	public ExtendedProperties(ClassLoader loader) {		
		this.data 		 = new PropertyData(loader);		
	}		
	public ExtendedProperties(PropertyData data) {				
		this.data		 = data;
	}
	
	public PropertyData getData() { return data; }
	
	/**
	 * the designated method to read properties from an inputstream. do not
	 * use the load method of java.util.Properties as this method may only
	 * throw a IOException
	 * @param stream InputStream from which the properties msut be loaded
	 * @throws IOException 
	 * @throws PropertyException
	 */
	public void read(InputStream stream) throws IOException, PropertyException {
		try {
			this.load(stream);
		} catch(PropertyRuntimeHelperException e) {
			throw e.getTargetException();
		}
	}
	
	/**
	 * this method is called during parsing of the input file. It's the
	 * right place to check if enhanced features need to be processed
	 * 
	 * @see java.util.Dictionary#put(java.lang.Object, java.lang.Object)
	 */
	public synchronized Object put(Object key, Object value) {
		if (value == null || key == null) return null;
		
		Property p = new Property(key.toString(), value.toString());
		try {
			if (data.getProcessor().processProperty(getData(), p)==false)
				return null;
		} catch (PropertyException e1) {
			throw new PropertyRuntimeHelperException(e1);
		}
		key 	= p.getKey();
		value	= p.getValue();
		
		addPropertyNode(p);
												
		this.data.setProperty(key.toString(), value.toString());
				
		return null;
	}
	
	private void addPropertyNode(Property p) {						
		String key            = p.getKey();
		String lastPart       = null;
		
		if (key.indexOf('.')>=0) {
			lastPart = key.substring(key.lastIndexOf('.')+1);
			key      = key.substring(0, key.lastIndexOf('.'));
		} else {
			return;
		}

		Object obj = this.data.getObjectProperty(key);
		List keys  = null;
		if (obj != null) {
			if (obj instanceof List) {
				keys = (List)obj;
			} else {
				throw new PropertyInitException("Inconsistent property structure key:" + key + " obj: " + obj.toString(), null);
			}			
		} else {
			keys = new ArrayList();
			this.data.setObjectProperty(key, keys);
		}		
		keys.add(lastPart);
		
	}
	/**
	 * @see java.util.Properties#getProperty(java.lang.String)
	 */
	public String getProperty(String key) {
		return (String)data.getProperty(key);
	}
}

class PropertyRuntimeHelperException extends RuntimeException {
	private PropertyException e;
	
	public PropertyRuntimeHelperException(PropertyException e) {
		this.e = e;
	}
	public PropertyException getTargetException() {
		return e;
	}
}
