/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on 25.02.2004
 */
package at.spardat.properties;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import at.spardat.properties.processor.*;

/**
 * Encapsulates the property data and attribute information.
 * Properties are reloaded during access to a property. No seperate thread is used.
 *
 * define the reload interval by setting the property "at.spardat.properties.ReloadInterval=<i>seconds</i>"
 *
 * @author Martin Kahr
 */
public class PropertyData {
	private static final String 		PROPERTY_RELOAD_INTERVAL = "at.spardat.properties.ReloadInterval";
	private static PropertyData 		INSTANCE;

	private Map							properties;
	private PropertyAttributes 			attributes;		//metadaten
	private ClassLoader 			  	classLoader;
	private AbstractPropertyProcessor 	processor;
	private Stack						loadPathStack;
	private long						readTimeStamp;
	private long						reloadInterval = -1;
    private boolean                     reload = true;

	public PropertyData() {
		this(PropertyData.class.getClassLoader());
	}
	public PropertyData(ClassLoader classLoader) {
		this.classLoader   = classLoader;
		this.properties    = new HashMap();
		this.attributes    = new PropertyAttributes();
		this.readTimeStamp = System.currentTimeMillis();
		loadPathStack      = new Stack();
		addToLoadPathStack("");

		this.processor   =  new VariableProcessor();
		MetaProcessor meta = new MetaProcessor();
		PraefixProcessor pre = new PraefixProcessor();
		IncludeProcessor inc = new IncludeProcessor();
		ConditionProcessor cond = new ConditionProcessor();
		PostfixProcessor post = new PostfixProcessor();	//6. Prozessor
		inc.setNextProcessor(post);			//5. Prozessor
		cond.setNextProcessor(inc);			//4. Prozessor
		pre.setNextProcessor(cond);			//3. Prozessor
		meta.setNextProcessor(pre);			//2. Prozessor
		processor.setNextProcessor(meta);	//1. Prozessor
	}

	public String getCurrentLoadPath() {
		return (String)loadPathStack.peek();
	}
	public void addToLoadPathStack(String path) {
		this.loadPathStack.push(path);
	}
	public void removeFromLoadPathStack() {
		this.loadPathStack.pop();
	}

	public AbstractPropertyProcessor getProcessor() { return processor; }
	public ClassLoader getClassLoader() { return classLoader; }

	public PropertyAttributes getAttributes() {
		return attributes;
	}

	public String getPropertyOrSystemProperty(String key) {
		String value = getProperty(key);
		if (value == null)
			return System.getProperty(key);
		return value;
	}
	public String getProperty(String key) {
		Object value = getObjectProperty(key);
		if (value instanceof String)
			return (String)value;
		else
			return null;
	}
	public Object getObjectProperty(String key) {
		return properties.get(key);
	}
	public void setObjectProperty(String key, Object value) {
		this.properties.put(key, value);
	}
	public void setProperty(String key, String value) {
		if (key.equals(PROPERTY_RELOAD_INTERVAL)) {
			try {
				reloadInterval  = Long.parseLong(value);
				reloadInterval *= 1000; // to milliseconds
			} catch(NumberFormatException e) {
				System.err.println("Invalid value '" + value + "' for property " + PROPERTY_RELOAD_INTERVAL);
			}
		}
		this.properties.put(key, value);
	}
	public void clearProperties() {
		this.properties.clear();
	}

	/**
	 * reloads all properties
	 */
	public final void reload() {
		if (reload == false || reloadInterval == -1 || reloadInterval > (System.currentTimeMillis()-readTimeStamp)) {
			return;
		}
		// mark this instance as not reloadable any more - this instance
		// will be replaced by the forceReload!!
		reload = false;
		try {
		    forceReload();
        } catch(RuntimeException ex) {
            // damit wieder ein reload probiert wird
            reload = true;
            this.readTimeStamp = System.currentTimeMillis();
            //throw ex;
        }
	}

	/**
	 * forces a reload
	 */
	public final void forceReload() {
		// TODO: use the correct root filename
        PropertyData data = PropertyLoader.loadRoot(this.getClassLoader());
		synchronized(PropertyData.class) {
            INSTANCE = data;
		}
	}

	public static PropertyData getInstance() {
		if (INSTANCE == null) {
			synchronized(PropertyData.class) {
				try {
                    INSTANCE = PropertyLoader.loadRoot();
                } catch(RuntimeException ex) {
                    ex.printStackTrace();
                    throw ex;
                }
			}
		} else {
		    return INSTANCE;
        }
		return INSTANCE;
	}

	public static void reset() {
		INSTANCE = null;
	}


	public Map getPropertiesMap(){
		return properties;
	}
}
