/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * @(#) $$Id: XProperties.java 2091 2007-11-28 14:14:59Z s3460 $$
 *
 *
 */
package at.spardat.properties;

import java.io.Serializable;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Root Class for accessing Properties. XProperties instances are the primary
 * way to access properties. The properties are loaded automatically on first access.
 * 
 * Use the static methods to get a concrete XProperties instance. 
 * The properties shall be defined in "packages" similary to java classes.
 * 
 * Each XProperties instance is initialized with a package name. If you
 * definie a property named "at.spardat.sample.a=0", you shall instantiate
 * a XProperties instance by calling <code>node=XProperties.getNode(Sample.class)</code> 
 * and access the property by calling <code>node.get("a")</code>
 * 
 * @author Martin Kahr
 */
public class XProperties implements Serializable {

	private String baseName;

	
	/** 
     * get the preferences node for the package of the given class
     * @param clazz The class where the packagename defines the base name for
     * property access
     * @return a XProperties instance which referes to the properties of the package of the given class  
     */ 
	public static XProperties getNode(Class clazz) {
		if (clazz == null) throw new IllegalArgumentException("Class parameter is null");
		String name = clazz.getName();
		int idx = name.lastIndexOf('.');
		if (idx < 0) 
			return getNodeOfPackage("");
		else
			return getNodeOfPackage(name.substring(0, idx));
	}
	
    /**
     * Get a new XProperties instance for a subnode of this node.  
     * @param node the name of the subnode
     * @return a new XProperties instance which referes to the subnode
     */
	public XProperties getSubNode(String node) {
		if(node == null || node.length() < 1) throw new IllegalArgumentException("Node is null or length is 0");
		return getNodeOfPackage(baseName + node);
	}
	
	/**
	 * get the root preferences node (no associated package)
     * @return root node of the preferences	 
	 */
	public static XProperties getRoot() {
		return getNodeOfPackage(null);		
	}
	
	/** 
	 * get the preferences node for the given package
	 * This method should not be used extensivly as it breaks some
	 * sort of encapsulation. This may lead to debugging sessions in case
	 * a package is refactored. 
	 * The getNode() package shall be prefered for access to properties. Static
	 * methods in some class of the target package shall be provided to 
	 * access the properties of another package!! 
     * @param packageName name of the package the XProperties shall refer to
     * @return XProperties instance which referes to the properties of the given package 
	 */ 
	public static XProperties getNodeOfPackage(String packageName) {
		return new XProperties(packageName);	
	}
	
	/**
	 * @deprecated for compatibility to Preferences API
     * @see at.spardat.properties.XProperties#getNode
     * @param clazz The class where the packagename defines the base name for
     * property access
     * @return a XProperties instance which referes to the properties of the package of the given class
	 */
	public static XProperties userNodeForPackage(Class clazz) {
		return getNode(clazz);
	}
	
	/**
	 * @return the name of the package this object references to
	 */
	public String getBaseName() {
		if (baseName.endsWith("."))
			return baseName.substring(0, baseName.length()-1);
		return baseName;
	}
	
	/**
     * TODO KDM: hier pruefen ob richtig initialisiert wurde
     * 
     * XProperties constructor. Access only through class methods.
     * @param value basename this instance shall refer to  
     */ 
	protected XProperties(String value) {
		this.baseName = value;
		if (baseName != null && baseName.length()>0) 
			this.baseName += "."; 
		else 
			this.baseName = "";		
	}
	
    /**
     * Internal method to access a raw property value
     * @param key the key of this property without the basename
     * @return The found property or null if no property was found
     */
	protected String getProperty(String key) {
		PropertyData.getInstance().reload();
		return PropertyData.getInstance().getProperty(baseName + key);
	}
    /**
     * Internal method to access a raw property value
     * @param key the key of this property without the basename
     * @return The found property 
     * @throws PropertyUnknownException if no property with the given name was found
     */
	protected String getMustProperty(String key) throws PropertyUnknownException {
		String value = getProperty(key);
		if (value == null)
			throw new PropertyUnknownException(key);
		return value;
	}

    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned.
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @return the property value
     */
	public String get(String key, String defaultValue) {
		String value = getProperty(key);
		return value != null ? value : defaultValue; 
	}

    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>int</code> value.   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @throws java.lang.NumberFormatException if the property may not be converted to an int
     * @return the property value
     */    
	public int getInt(String key, int defaultValue) {
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return Integer.parseInt(value);	 
	}
    
    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>float</code> value.   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @throws java.lang.NumberFormatException if the property may not be converted to an float
     * @return the property value
     */        
	public float getFloat(String key, float defaultValue) {
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return Float.parseFloat(value);	 		
	}
    
    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>double</code> value.   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @throws java.lang.NumberFormatException if the property may not be converted to an double
     * @return the property value
     */        
	public double getDouble(String key, double defaultValue) {
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return Double.parseDouble(value);	 		
	}
    
    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>long</code> value.   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @throws java.lang.NumberFormatException if the property may not be converted to an long
     * @return the property value
     */        
	public long getLong(String key, long defaultValue) {
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return Long.parseLong(value);	 		
	}	
    
    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>boolean</code> value.   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @return the property value
     */        
	public boolean getBoolean(String key, boolean defaultValue) {
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return Boolean.valueOf(value).booleanValue();	 		
	}
	
    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>java.util.Date</code> by using the formating pattern "DD.MM.YYYY".   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @throws ParseException if the conversion fails
     * @return the property value
     */                
	public Date getDate(String key, Date defaultValue) throws ParseException {
		DateFormat dateFormatter = new SimpleDateFormat("dd.MM.yyyy");
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return dateFormatter.parse(value);				
	}
	
    /**
     * Access to the property with the name "basename+key". If no property
     * was found the defaultvalue is returned. The found property value is converted
     * to an <code>java.util.Date</code> by using the formating pattern "HH:MM:SS".   
     * @param key the name of the property
     * @param defaultValue the defaultvalue if no property is defined
     * @throws ParseException if the conversion fails
     * @return the property value
     */                    
	public Date getTime(String key, Date defaultValue) throws ParseException {
		DateFormat timeFormatter = new SimpleDateFormat("HH:mm:ss");
		String value = getProperty(key);
		if (value == null)
			return defaultValue;
		else
			return timeFormatter.parse(value);
	}
	
    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown.
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found. 
     * @return the property value
     */
	public String get(String key) throws PropertyUnknownException {
		return getMustProperty(key);
	}

    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown.
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws NumberFormatException if the property conversion to int fails 
     * @return the property value
     */    
	public int getInt(String key)  throws PropertyUnknownException {
		return Integer.parseInt(getMustProperty(key));	 
	}
    
    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown.
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws NumberFormatException if the property conversion to float fails 
     * @return the property value
     */        
	public float getFloat(String key)  throws PropertyUnknownException {
		return Float.parseFloat(getMustProperty(key));
	}
    
    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown.
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws NumberFormatException if the property conversion to double fails 
     * @return the property value
     */        
	public double getDouble(String key)  throws PropertyUnknownException {
		return Double.parseDouble(getMustProperty(key));
	}
    
    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown.
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws NumberFormatException if the property conversion to long fails 
     * @return the property value
     */        
	public long getLong(String key)  throws PropertyUnknownException {
		return Long.parseLong(getMustProperty(key));
	}	
    
    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown.
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws NumberFormatException if the property conversion to boolean fails 
     * @return the property value
     */        
	public boolean getBoolean(String key)  throws PropertyUnknownException {
		return Boolean.valueOf(getMustProperty(key)).booleanValue();
	}

    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown. The property is converted to a "java.util.Date" by
     * using the formating pattern "DD.MM.YYYY".
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws ParseException if the property conversion fails 
     * @return the property value
     */        
	public Date getDate(String key) throws PropertyUnknownException, ParseException {
		DateFormat dateFormatter = new SimpleDateFormat("dd.MM.yyyy");
		return dateFormatter.parse(getMustProperty(key));				
	}
	
    /**
     * Access to the property with the name "basename+key". If no property
     * an exception is thrown. The property is converted to a "java.util.Date" by
     * using the formating pattern "HH:MM:SS".
     * @param key the name of the property
     * @throws PropertyUnknownException if no property was found.
     * @throws ParseException if the property conversion fails 
     * @return the property value
     */            
	public Date getTime(String key) throws PropertyUnknownException, ParseException {
		DateFormat timeFormatter = new SimpleDateFormat("HH:mm:ss");
		return timeFormatter.parse(getMustProperty(key));
	}

	/**	 
     * Returns an Iterator to iterate over the names of the properties which
     * are defined on this node. Note: This does not include any subnodes which
     * are defined on this node!
	 * @return an Iterator with String elements representing the names of defined properties   
	 */
	public Iterator getKeys() {
		String nodeName= baseName;
		if (nodeName.endsWith(".")) nodeName = nodeName.substring(0, nodeName.length()-1);
		Object keyList = PropertyData.getInstance().getObjectProperty(nodeName);
		if (keyList == null || !(keyList instanceof List)) return new EmptyIterator();
		return ((List)keyList).iterator();
	}
	
	/**
     * Tries to find subnodes of this XProperties node which start with name <code>propertyName</code>
     * and a number starting with "1" (e.g. at.spardat.sample.group1).
     * The returned Iterator contains an XProperties instance for each found subnode. 
     * 
	 * @see at.spardat.properties.test.XPropertiesTest#testGroupProperties() for a sample
	 * @param propertyName the basename of the subnode without the counter
	 * @return an iterator containing a number of XProperties instance for each found group element
	 */	
	public Iterator getGroupProperties(String propertyName) {
		return getGroupProperties(propertyName, 1);
	}
    
    /**
     * Tries to find subnodes of this XProperties node which start with name <code>propertyName</code>
     * and a number starting with <code>start</code> (e.g. at.spardat.sample.group1).
     * The returned Iterator contains an XProperties instance for each found subnode. 
     * 
     * @see at.spardat.properties.test.XPropertiesTest#testGroupProperties() for a sample
     * @param propertyName the basename of the subnode without the counter
     * @param start the start number for the counter
     * @return an iterator containing a number of XProperties instance for each found group element
     */     
	public Iterator getGroupProperties(String propertyName, int start) {
		ArrayList groups = new ArrayList();
				
		String propName  = propertyName + start;
		
		XProperties props= this.getSubNode(propName);
		Iterator keys    = props.getKeys();
		while(props != null && !(keys instanceof EmptyIterator)) {
			groups.add(props);
			propName = propertyName + (++start);			
			props    = this.getSubNode(propName);
			keys     = props.getKeys();
		}
		return groups.iterator();	
	}

	private static List processIter(Iterator iter,Map myMap,List keyList){
		PropertyData.getInstance().reload();
		while (iter.hasNext()){
			String tmpKey = (String) iter.next();			
			Object tmpValue = myMap.get(tmpKey);
			if (tmpValue instanceof String) {
				keyList.add(tmpKey);
			}if (tmpValue instanceof List) {
				processIter(((List)tmpValue).iterator(),myMap,keyList);
			}
		}
		return keyList;
	}

	/**Liefert zum bergebenen PropertyKey die diesbezglich konfigurierten Informationen.
	 * @param prop
	 * @return
	 */
	public static String getPropertyMeta(String prop){
		StringBuffer sb = new StringBuffer();
		if (PropertyData.getInstance().getAttributes().getAttribute(prop) != null){
			Iterator iter = PropertyData.getInstance().getAttributes().getAttribute(prop).getPropertiesWithAttributeValue(prop);
			while (iter!= null && iter.hasNext()){
				sb.append(iter.next());
			}
		}
		return sb.toString();
	}
	
	/**Liefert alle vorhandenen PropertyKeys
	 * @return
	 */
	public static Iterator getAllKeys(){
		List keyList = new ArrayList();
		Map myMap = PropertyData.getInstance().getPropertiesMap();
		keyList = processIter(myMap.keySet().iterator(),myMap,keyList);
		Collections.sort(keyList);
		return keyList.iterator();
	}
}

class EmptyIterator implements Iterator {
	public boolean hasNext() {
		return false;
	}
	public Object next() {
		return null;
	}
	public void remove() {
	}
}
