/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * Created on 25.02.2004
 */
package at.spardat.properties.processor;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.StringTokenizer;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import at.spardat.properties.*;
import at.spardat.properties.ExtendedProperties;
import at.spardat.properties.Property;
import at.spardat.properties.PropertyData;
import at.spardat.properties.PropertyException;
import at.spardat.properties.PropertyFileNotFoundException;

/**
 * this processor is responsible for handling the include of property files with
 * the syntax "@include=path_to_property_file"
 * 
 * @author Martin Kahr
 */
public class IncludeProcessor extends AbstractPropertyProcessor {

    public final static String INCLUDE_PRAEFIX = "@include";

    /**
     * @see at.spardat.properties.AbstractPropertyProcessor#doProcessProperty(at.spardat.properties.PropertyData,
     *      at.spardat.properties.Property)
     */
    protected boolean doProcessProperty(PropertyData data, Property property)
            throws PropertyException {
        if (property.getKey().startsWith(INCLUDE_PRAEFIX)) {
            try {
                includePropertyFile(data, property.getValue().toString());
            } catch (IOException e) {
                e.printStackTrace();
                throw new PropertyFileNotFoundException(property.getValue().toString());
            }
            return false;
        }
        return true;
    }

    private void includeFromClassloader(PropertyData data, String fileName)
            throws IOException, PropertyException {
        if (fileName.startsWith("/")) {
            fileName = fileName.substring(1);
        } else {
            fileName = data.getCurrentLoadPath() + fileName;
        }
        String newLoadPath = "";
        int idx = fileName.lastIndexOf('/');
        if (idx >= 0) {
            newLoadPath = fileName.substring(0, idx + 1);
        }
        ExtendedProperties included = new ExtendedProperties(data);
        data.addToLoadPathStack(newLoadPath);
        InputStream inStream = null;
        try {
            inStream = data.getClassLoader().getResourceAsStream(fileName);
            if (inStream == null)
                throw new PropertyFileNotFoundException(fileName);
            included.read(inStream);
        } finally {
            if (inStream != null)
                try {
                    inStream.close();
                } catch (Exception ignored) {
                }
            data.removeFromLoadPathStack();
        }
    }

    private void includeFromFile(PropertyData data, String fileName) throws IOException,
            PropertyException {
        URL fileURL = null;
        try {
            fileURL = new URL(fileName);
        } catch (MalformedURLException e) {
            throw new PropertySyntaxException("Malformed URI for file import. "
                    + fileName);
        }
        File file = new File(fileURL.getFile());
        if (file.exists() == false)
            throw new PropertyFileNotFoundException(fileName);

        FileInputStream fIn = null;
        try {
            fIn = new FileInputStream(file);
            ExtendedProperties included = new ExtendedProperties(data);
            included.read(fIn);
        } finally {
            if (fIn != null)
                try {
                    fIn.close();
                } catch (Exception ignored) {
                }
        }
    }

    /**
     * Reads Properties from a JDCB data source
     * @param data
     * @param fileName
     * @throws IOException
     * @throws PropertyException
     * @since version_number
     * @author s3460
     */
    private void includeFromDataSource(PropertyData data, String fileName)
            throws IOException, PropertyException {
        
        if (fileName.indexOf('/') < 0) {
        //correct form: @include datasource:datasource_name/table_name
            throw new PropertySyntaxException(
                    "@include for datasource import contains no '/': " + fileName);
        }
        //Parse @include
        String dsName = fileName.substring(fileName.indexOf(":") + 1, fileName
                .indexOf('/'));
        String tableName = fileName.substring(fileName.indexOf("/") + 1);
        //get Connection
        Connection con = getPoolConnection(dsName);       
        
        executeDBQuery(data,con,tableName);        
    }

    /**
     * Reads Properties from a JDCB direct DB connection
     * @param data
     * @param fileName
     * @throws IOException
     * @throws PropertyException
     * @since version_number
     * @author s3460
     */
    private void includeFromJDBC(PropertyData data, String fileName)
    throws IOException, PropertyException {
        String driver = null;
        String url = null;
        String user = null;
        String pw = null;
        String tableName = null;
        //Parse String: 'jdbc:driver/url/user/pw/table'
        String dbinfo = fileName.substring(fileName.indexOf(":")+1);        
        StringTokenizer st = new StringTokenizer(dbinfo,"/");
        int i = 0;
        for (;st.hasMoreTokens();i++) {
            if(i==0)
                driver = st.nextToken();
            if(i==1)
                url = st.nextToken();
            if(i==2)
                user = st.nextToken();
            if(i==3)
                pw = st.nextToken();
            if(i==4)
                tableName = st.nextToken();
        }

//      correct form: 
		// jdbc:oracle.jdbc.OracleDriver/jdbc:oracle:thin:@datacashd.server.lan.at:1521:CASHD1DB/user/pw/table_name
		if (i < 5) {		
		    throw new PropertySyntaxException(
		            "@include for jdbc import not in form 'jdbc:driver/url/user/pw/table': " + fileName);
		}		
		
		Connection con = getJDBCConnection(driver,url,user,pw);     		
		executeDBQuery(data,con,tableName);        
	}    
    
    /**
     * Get Connection by DriverManager.
     * @param driver
     * @param url
     * @param user
     * @param pw
     * @return
     * @since version_number
     * @author s3460
     */
    private Connection getJDBCConnection(String driver, String url, String user, String pw) {
        Connection con = null;
        try {
            Class.forName(driver);
            con =  DriverManager.getConnection(url, user, pw);
        } catch (ClassNotFoundException e) {
            throw new RuntimeException("JDBC Driver not Found:" + driver,e);
        } catch (SQLException e) {
            throw new RuntimeException("Error opening Connection to " + url,e);
        }
        
        return con;
    }

    /**
     * Executes Query to retrieve the Properties into 'data' and closes the Connection.
     * @param data
     * @param con
     * @param tableName
     * @since version_number
     * @author s3460
     */
    private void executeDBQuery(PropertyData data, Connection con, String tableName) {
        PreparedStatement selectStmt = null;
        ResultSet rs = null;        
        //Abfragen DB and fill data
        try {
            selectStmt = con.prepareStatement("SELECT TXT_KEY, TXT_VALUE FROM " + tableName);            
            rs = selectStmt.executeQuery();
            while (rs.next()) {
                String key = rs.getString("TXT_KEY");
                String value = rs.getString("TXT_VALUE");
                data.setProperty(key, value);
            }
        } 
        catch (SQLException e) {
            throw new RuntimeException("SQLException on 'SELECT TXT_KEY, TXT_VALUE FROM' " + tableName,e);
        }
        finally {
            try {
                if (rs != null)
                    rs.close();
            } catch (SQLException e2) {                
            }
            
            try {
                if (selectStmt != null)
                    selectStmt.close();
            } catch (SQLException e1) {                
            }
           
            try {
                if (con != null)
                    con.close();
            } catch (SQLException e3) {                
            }
        }        
    }

    /**
     * Get Connection by JNDI lookup.
     * @param poolName
     * @return
     * @since version_number
     * @author s3460
     */
    private Connection getPoolConnection(String poolName) {
        Context ctx = null;

        try {
            DataSource ds = null;
            // get new datasource
            ctx = new InitialContext();
            try {
                ds = (DataSource) ctx.lookup("java:comp/env/jdbc/" + poolName);
            } catch (NamingException ex) {
            }
            if (ds == null) {
                // hope this is the correct name
                ds = (DataSource) ctx.lookup(poolName);
            }

            return ds.getConnection();

        } catch (NamingException e) {
            throw new RuntimeException("DataSource not found: " + poolName,e);
        } catch (SQLException e) {
            throw new RuntimeException("SQLException on open DataSource: " + poolName,e);
        } finally {
            if (ctx != null)
                try {
                    ctx.close();
                } catch (Exception ignore) {
                }
        }
    }

    private void includePropertyFile(PropertyData data, String fileName)
            throws IOException, PropertyException {
        if (fileName.startsWith("file:")) {
            includeFromFile(data, fileName);
        } else if (fileName.startsWith("datasource:")) {
            includeFromDataSource(data, fileName);
        } else if (fileName.startsWith("jdbc:")) {
            includeFromJDBC(data, fileName);
        } else {
            includeFromClassloader(data, fileName);
        }
    }
}
